/*
 * EXEPATH - returns the directory from which the program was loaded.
 *
 * This file is part of the Mateusz' DOS Routines (MDR): http://mdr.osdn.io
 * Published under the terms of the MIT License, as stated below.
 *
 * Copyright (C) 2014-2022 Mateusz Viste
 *
 * Permission is hereby granted, free of charge, to any person obtaining a copy
 * of this software and associated documentation files (the "Software"), to
 * deal in the Software without restriction, including without limitation the
 * rights to use, copy, modify, merge, publish, distribute, sublicense, and/or
 * sell copies of the Software, and to permit persons to whom the Software is
 * furnished to do so, subject to the following conditions:
 *
 * The above copyright notice and this permission notice shall be included in
 * all copies or substantial portions of the Software.
 *
 * THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
 * IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
 * FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
 * AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
 * LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING
 * FROM, OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS
 * IN THE SOFTWARE.
 */

/*** Excerpt from TechHelp! **************************************************
 Starting with DOS 3.0, DOS stores an additional string after the end of the
 formal environment that can be used to locate the drive and directory from
 which a program was loaded. Applications can use this to help locate program
 overlays and data files.

 This becaem important with DOS 3.0, when DOS began allowing an operator to
 invoke a program by prefixing a pathname before the program filename.

 Following the final string of the environment is a byte of 0, indicating the
 end of the formal environment. The next two bytes are a 16-bit binary count
 of additional strings (this value is normally 0001H). Following this binary
 value, you can expect to find an ASCIIZ string of the filespec (including the
 drive, path, and extension) which was used by DOS Fn 4bH (EXEC) to load and
 execute the program.

 Even if the program file was not in the default directory and DOS or Windows
 needed to use the command search path to locate the program file, the correct
 full filespec (d:\path\filename.ext,0) will be stored after the end of the
 environment.
*****************************************************************************/

#include <i86.h> /* MK_FP() */

#include <mdr\dos.h>


/* fetches directory where the program was loaded from and return its length.
 * path string is never longer than 128 (incl. the null terminator) and it is
 * always terminated with a backslash separator, unless it is an empty string */
unsigned char mdr_dos_exepath(char *path) {
  char far *env;
  unsigned int i;
  int lastsep;

  /* compute the env pointer */
  env = mdr_dos_env();

  /* skip all environment variables */
  for (;;) {
    while (*env != 0) env++; /* go to end of variable */
    env++;
    if (*env == 0) break; /* end of list */
  }

  /* read the WORD that indicates the amount of strings that follow - must
   * be at least 1 since the full prog name is the first one */
  env++;
  if (*((unsigned short far *)env) == 0) {
    path[0] = 0;
    return(0);
  }

  /* copy the EXEPATH to result and truncate after last '\' */
  env += 2;
  lastsep = -1;
  for (i = 0;; i++) {
    path[i] = *env;
    if (path[i] == '\\') lastsep = i;
    if (path[i] == 0) break; /* end of string */
    if (i >= 126) {
      lastsep = -1;
      break;     /* this DOS string should never go beyond 127 chars! */
    }
    env++;
  }
  lastsep++;
  path[lastsep] = 0;
  return(lastsep);
}
